#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os

# pathlib.Path should not be used to keep consistency with str paths
import pathlib

import octobot_commons.enums as enums
import octobot_commons.logging as logging
import octobot_commons.constants as constants
import octobot_commons.databases.run_databases.run_databases_identifier as run_databases_identifier


def get_backtesting_related_run_path_identifiers_str(database_adaptor):
    """
    :return: database identifier fragments associated to the given database_adaptor used
    in backtesting
    """
    separator = (
        os.path.sep if database_adaptor.is_file_system_based else constants.DB_SEPARATOR
    )
    return {
        f"{separator}{enums.RunDatabases.BACKTESTING.value}{separator}",
        f"{separator}{enums.RunDatabases.OPTIMIZER.value}{separator}",
    }


def get_global_run_database_identifier(runs_identifier):
    """
    :return: a RunDatabasesIdentifier associated to the given runs_identifier
    """
    # used to split paths into str parts
    split_path = pathlib.Path(runs_identifier).parts
    try:
        if split_path[-2] == enums.RunDatabases.OPTIMIZER.value:
            # in optimizer
            # ex: [..., 'DipAnalyserTradingMode', 'Dip Analyser strat designer test', 'optimizer', 'optimizer_1']
            optimizer_id = (
                run_databases_identifier.RunDatabasesIdentifier.parse_optimizer_id(
                    split_path[-1]
                )
            )
            campaign_name = split_path[-3]
            trading_mode = split_path[-4]
            return run_databases_identifier.RunDatabasesIdentifier(
                trading_mode,
                optimization_campaign_name=campaign_name,
                backtesting_id=0,
                optimizer_id=optimizer_id,
            )
        # in backtesting
        # ex: [..., 'DipAnalyserTradingMode', 'Dip Analyser strat designer test', 'backtesting']
        campaign_name = split_path[-2]
        trading_mode = split_path[-3]
        return run_databases_identifier.RunDatabasesIdentifier(
            trading_mode,
            optimization_campaign_name=campaign_name,
            backtesting_id=0,
        )
    except IndexError as err:
        logging.get_logger("run_databases_utils").exception(
            err, True, f"Unhandled backtesting data path format: {runs_identifier}"
        )
        return None



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import octobot_commons.constants as commons_constants
import octobot_commons.enums as commons_enums
import octobot_commons.singleton as singleton
import octobot_commons.logging as logging
import octobot_commons.async_job as async_job
import octobot_commons.os_util as os_util


class ClockSynchronizer(singleton.Singleton):
    DEFAULT_SYNC_REFRESH_INTERVAL = (
        commons_constants.CLOCK_REFRESH_HOURS_INTERVAL
        * commons_constants.HOURS_TO_SECONDS
    )

    def __init__(self):
        super().__init__()
        self.sync_job = None
        self.sync_interval = self.DEFAULT_SYNC_REFRESH_INTERVAL
        self.logger = logging.get_logger(self.__class__.__name__)

    def _get_sync_cmd(self):
        platform = os_util.get_os()
        if platform is commons_enums.PlatformsName.WINDOWS:
            # use 2x w32tm /resync as the 1st one often fails
            return "net stop w32time && net start w32time && w32tm /resync & w32tm /resync && w32tm /query /status"
        if platform is commons_enums.PlatformsName.LINUX:
            return "sudo service ntp stop && sudo ntpd -gq && sudo service ntp start"
        if platform is commons_enums.PlatformsName.MAC:
            raise NotImplementedError
        raise NotImplementedError

    async def _sync_clock(self):
        proc = await asyncio.create_subprocess_shell(
            self._get_sync_cmd(),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout, stderr = await proc.communicate()
        if proc.returncode == 0:
            self.logger.info("Successful os clock synchronization")
        else:
            self.logger.warning(
                f"Error: Time synchronization command exited with {proc.returncode}] "
                f'command: "{self._get_sync_cmd()}"'
            )
        if stdout:
            self.logger.debug(f"[stdout] {stdout}")
        if stderr:
            self.logger.debug(f"[stderr] {stderr}")

    async def start(self) -> bool:
        """
        Start the clock synchronization loop if possible on this system
        :return: True if the loop has been started
        """
        try:
            self._get_sync_cmd()
        except NotImplementedError:
            self.logger.debug(
                "Disable clock synchronizer: not implemented on this system."
            )
            return False
        if os_util.has_admin_rights():
            self.logger.debug("Starting clock synchronizer")
            self.sync_job = async_job.AsyncJob(
                self._sync_clock, execution_interval_delay=self.sync_interval
            )
            await self.sync_job.run()
            return True
        self.logger.debug(
            "Clock synch loop disabled: admin rights are required to synchronize the computer clock"
        )
        return False

    def stop(self):
        """
        Stop the synchronization loop
        """
        self.logger.debug("Stopping clock synchronizer")
        if self.sync_job is not None and not self.sync_job.should_stop:
            self.sync_job.stop()


async def start_clock_synchronizer():
    """
    Start the clock synchronization loop if possible on this system
    :return: True if the loop has been started
    """
    return await ClockSynchronizer.instance().start()


async def stop_clock_synchronizer():
    """
    Stop the synchronization loop
    """
    return ClockSynchronizer.instance().stop()



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import time
from datetime import datetime

_EPOCH = time.time()
TIMEZONE_DELTA = datetime.fromtimestamp(_EPOCH) - datetime.utcfromtimestamp(_EPOCH)


def convert_timestamp_to_datetime(
    timestamp, time_format="%d/%m/%y %H:%M", force_timezone=False
):
    """
    Convert a timestamp to a datetime object
    :param timestamp: the timestamp to convert
    :param time_format: the time format
    :param force_timezone: if the timezone should be forced
    :return: the created datetime object
    """
    if force_timezone:
        timestamp += TIMEZONE_DELTA.seconds
    return datetime.fromtimestamp(timestamp).strftime(time_format)


def convert_timestamps_to_datetime(
    timestamps, time_format="%d/%m/%y %H:%M", force_timezone=False
):
    """
    Convert multiple timestamps to datetime objects
    :param timestamps: the timestamp to convert list
    :param time_format: the time format
    :param force_timezone: if the timezone should be forced
    :return: the created datetime objects
    """
    return [
        convert_timestamp_to_datetime(
            timestamp, time_format=time_format, force_timezone=force_timezone
        )
        for timestamp in timestamps
    ]


def is_valid_timestamp(timestamp):
    """
    Check if the timestamp is valid
    :param timestamp: the timestamp to check
    :return: the check result
    """
    if timestamp:
        try:
            datetime.fromtimestamp(timestamp)
        except OSError:
            return False
        except ValueError:
            return False
        except OverflowError:
            return False
        except TypeError:
            return False
    return True


def get_now_time(time_format="%Y-%m-%d %H:%M:%S"):
    """
    Get the current time
    :param time_format: the time format
    :return: the current timestamp
    """
    return datetime.fromtimestamp(time.time()).strftime(time_format)


def datetime_to_timestamp(date_time_str: str, date_time_format: str) -> float:
    """
    Convert a datetime to timestamp
    :param date_time_str: the datetime string
    :param date_time_format: the datetime format
    :return: the timestamp
    """
    return time.mktime(
        create_datetime_from_string(date_time_str, date_time_format).timetuple()
    )


def create_datetime_from_string(date_time_str: str, date_time_format: str) -> datetime:
    """
    Convert a string to datetime
    :param date_time_str: the datetime string
    :param date_time_format: the datetime format
    :return: the converted datetime
    """
    return datetime.strptime(date_time_str, date_time_format)



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
import shutil
import zipfile
import contextlib
import mock
import pathlib
import octobot_commons.constants as constants
import octobot_commons.profiles as profiles
import octobot_commons.profiles.profile_sharing as profile_sharing
from octobot_commons.profiles.profile_sharing import _get_unique_profile_folder, _ensure_unique_profile_id, \
    _get_profile_name
import octobot_commons.tests.test_config as test_config

from tests.profiles import profile, get_profile_path


def test_export_profile(profile):
    export_path = "exported"
    exported_file = f"{export_path}.zip"
    tentacles_config = os.path.join(get_profile_path(), "tentacles_config.json")
    spec_tentacles_config = os.path.join(get_profile_path(), "specific_config")
    other_profile = os.path.join(test_config.TEST_FOLDER, "other_profile")
    with _cleaned_tentacles(export_path,
                            exported_file,
                            tentacles_config,
                            dir1=spec_tentacles_config,
                            dir2=other_profile):
        # create fake tentacles config
        shutil.copy(profile.config_file(), tentacles_config)
        os.mkdir(spec_tentacles_config)
        shutil.copy(profile.config_file(), os.path.join(spec_tentacles_config, "t1.json"))
        shutil.copy(profile.config_file(), os.path.join(spec_tentacles_config, "t2.json"))
        with mock.patch.object(os, "remove", mock.Mock()) as remove_mock:
            profiles.export_profile(profile, export_path)
            remove_mock.assert_not_called()
        profiles.export_profile(profile, export_path)
        assert os.path.isfile(exported_file)
        with zipfile.ZipFile(exported_file) as zipped:
            zipped.extractall(other_profile)
        # ensure all files got zipped
        for root, dirs, files in os.walk(profile.path):
            dir_path = os.path.join(other_profile, "specific_config") if "specific_config" in root else other_profile
            assert all(
                os.path.isfile(os.path.join(dir_path, f))
                for f in files
            )


def test_export_profile_with_existing_file(profile):
    export_path = "exported"
    exported_file = f"{export_path}.zip"
    tentacles_config = os.path.join(get_profile_path(), "tentacles_config.json")
    spec_tentacles_config = os.path.join(get_profile_path(), "specific_config")
    other_profile = os.path.join(test_config.TEST_FOLDER, "other_profile")
    with _cleaned_tentacles(export_path,
                            exported_file,
                            tentacles_config,
                            dir1=spec_tentacles_config,
                            dir2=other_profile):
        # create fake tentacles config
        shutil.copy(profile.config_file(), tentacles_config)
        os.mkdir(spec_tentacles_config)
        shutil.copy(profile.config_file(), os.path.join(spec_tentacles_config, "t1.json"))
        shutil.copy(profile.config_file(), os.path.join(spec_tentacles_config, "t2.json"))
        shutil.copy(profile.config_file(), f"{export_path}.{constants.PROFILE_EXPORT_FORMAT}")
        with mock.patch.object(os, "remove", mock.Mock()) as remove_mock:
            profiles.export_profile(profile, export_path)
            remove_mock.assert_called_once_with(f"{export_path}.{constants.PROFILE_EXPORT_FORMAT}")
        assert os.path.isfile(exported_file)
        with zipfile.ZipFile(exported_file) as zipped:
            zipped.extractall(other_profile)
        # ensure all files got zipped
        for root, dirs, files in os.walk(profile.path):
            dir_path = os.path.join(other_profile, "specific_config") if "specific_config" in root else other_profile
            assert all(
                os.path.isfile(os.path.join(dir_path, f))
                for f in files
            )


def test_import_install_profile(profile):
    export_path = os.path.join(test_config.TEST_FOLDER, "super_profile")
    exported_file = f"{export_path}.zip"
    spec_tentacles_config = os.path.join(get_profile_path(), "specific_config")
    tentacles_config = os.path.join(get_profile_path(), "tentacles_config.json")
    other_profile = os.path.join(constants.USER_PROFILES_FOLDER, "imported_super_profile")
    with _cleaned_tentacles(export_path,
                            exported_file,
                            tentacles_config,
                            dir1=other_profile,
                            dir2=constants.USER_FOLDER,
                            dir3=spec_tentacles_config):
        # create fake tentacles config
        shutil.copy(profile.config_file(), tentacles_config)
        os.mkdir(spec_tentacles_config)
        shutil.copy(profile.config_file(), os.path.join(spec_tentacles_config, "t1.json"))
        shutil.copy(profile.config_file(), os.path.join(spec_tentacles_config, "t2.json"))
        profiles.export_profile(profile, export_path)
        imported_profile_path = os.path.join(constants.USER_PROFILES_FOLDER, "imported_super_profile")
        with mock.patch.object(profile_sharing, "_ensure_unique_profile_id", mock.Mock()) \
                as _ensure_unique_profile_id_mock:
            imported_profile = profiles.import_profile(exported_file)
            assert isinstance(imported_profile, profiles.Profile)
            profile.read_config()
            assert profile.name == imported_profile.name
            assert profile.path != imported_profile.path
            assert profile.imported is False
            assert imported_profile.imported is True
            _ensure_unique_profile_id_mock.assert_called_once()
        assert os.path.isdir(imported_profile_path)
        # ensure all files got imported
        for root, dirs, files in os.walk(profile.path):
            dir_path = os.path.join(other_profile, "specific_config") if "specific_config" in root else other_profile
            assert all(
                os.path.isfile(os.path.join(dir_path, f))
                for f in files
            )
        assert isinstance(profiles.import_profile(exported_file), profiles.Profile)
        assert os.path.isdir(f"{imported_profile_path}_2")
        with mock.patch.object(shutil, "rmtree", mock.Mock()) as shutil_rmtree_mock:
            assert isinstance(profiles.install_profile(exported_file,
                                                       _get_profile_name(None, exported_file),
                                                       ".",
                                                       True,
                                                       False), profiles.Profile)
            shutil_rmtree_mock.assert_called_once()
        assert os.path.isdir(imported_profile_path)
        assert not os.path.isdir(f"{imported_profile_path}_3")


def test_get_unique_profile_folder(profile):
    assert _get_unique_profile_folder(profile.config_file()) == f"{profile.config_file()}_2"
    other_file = f"{profile.config_file()}_2"
    other_file_2 = f"{profile.config_file()}_3"
    other_file_3 = f"{profile.config_file()}_5"
    with _cleaned_tentacles(other_file, other_file_2, other_file_3):
        shutil.copy(profile.config_file(), other_file)
        assert _get_unique_profile_folder(profile.config_file()) == f"{profile.config_file()}_3"
        shutil.copy(profile.config_file(), other_file_2)
        assert _get_unique_profile_folder(profile.config_file()) == f"{profile.config_file()}_4"
        shutil.copy(profile.config_file(), other_file_3)
        assert _get_unique_profile_folder(profile.config_file()) == f"{profile.config_file()}_4"


def test_ensure_unique_profile_id(profile):
    other_profile = "second_profile"
    profiles_path = pathlib.Path(profile.path).parent
    other_profile_path = profiles_path.joinpath(other_profile)
    with _cleaned_tentacles(dir1=other_profile_path):
        shutil.copytree(profile.path, other_profile_path)
        other_profile = profiles.Profile(other_profile_path).read_config()
        _ensure_unique_profile_id(other_profile)
        ids = profiles.Profile.get_all_profiles_ids(profiles_path)
        assert len(ids) == 2
        # changed new profile id
        assert ids[0] != ids[1]


@contextlib.contextmanager
def _cleaned_tentacles(*items, **dirs):
    try:
        for item in items:
            if os.path.isfile(item):
                os.remove(item)
        for directory in dirs.values():
            if os.path.isdir(directory):
                shutil.rmtree(directory)
        yield
    finally:
        for item in items:
            if os.path.isfile(item):
                os.remove(item)
        for directory in dirs.values():
            if os.path.isdir(directory):
                shutil.rmtree(directory)
